#   Copyright 2020-2021 Exactpro (Exactpro Systems Limited)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from abc import ABC, abstractmethod
from typing import List

from google.protobuf.message import Message

from th2_common.schema.message.configuration.message_configuration import RouterFilterConfiguration


class FilterStrategy(ABC):

    @abstractmethod
    def verify(self, message: Message, router_filter: RouterFilterConfiguration = None,
               router_filters: List[RouterFilterConfiguration] = None):
        pass
